<?php
require_once('utils/bdUtils.php');

class oauthClient {
    protected $error;
    protected $access_token;
    protected $expires_in;
    protected $assertion_type;
    protected $access_token_secret;
    protected $resource;
    protected $url_rs;
    protected $scope;
    protected $bearer;
    protected $algorithm;
    protected $assertion;
    protected $isExpired;
    protected $deactiveDebug;
    protected $key;

    /**
     * Creates an instance of an OAuth Client
     * @param <type> assertion_type Type of the  assertion provided. Currently supported saml2 and PAPI assertions.
     */
    public function __construct($assertion_type) {
        $this->error = null;
        $this->access_token=null;
        $this->assertion_type = $assertion_type;
        $this->expires_in =null;
        $this->resource = null;
        $this->url_rs = null;
        $this->scope= null;
        $this->bearer=null;
        $this->algorithm=null;
        $this->assertion=null;
        $this->isExpired = false;
        $this->deactiveDebug();
        $this->key="xxxx";

    }


    /**
     * Manage all the steps needed to make an Access Token request.
     * @param String $assertion Assertion provided
     * @param String $scope URI of the scope of the resource
     * @param String $url_as URL of the OAuth authorization server
     * @param String $url_rs URL of the OAuth resource server
    * @param String lifetime OPTIONAL Lifetime of the access or bearer token. Default value: 3600 seconds
     */
    public function requestAccessToken($assertion,$scope, $url_as, $url_rs, $lifetime=null) {
        if($this->activeDebug) error_log("oauthClient.requestAccessToken");
        $at = null;
        $data = $this->generateATRequest($assertion, $scope, $lifetime);
        $this->url_rs=$url_rs;
        $this->scope = $scope;
        $this->assertion = $assertion;
        $this->postConnection($url_as, $data);
    }

    /**
     * Function that generates de access token requests.
     * @param <type> $assertion Assertion provided
     * @param <type> $scope URI of the scope of the resource
     * @param <type> $lifetime Lifetime of the access or bearer token    
     * @return Array with the attributes specified by the draft-ietf-oauth-v2-05  '3.10.1. Client Requests Access Token': type, assertion, scope, format, lifetime and response_format
     */
    private function generateATRequest($assertion, $scope, $lifetime) {
        if($this->activeDebug) error_log("oauthClient.generateATRequest");
        $array = array();
        $array['grant_type'] = "assertion";
        $array['assertion'] = urlencode(json_encode($this->transformArray($assertion)));
        $array['scope'] = $scope;
        $array['assertion_type'] = $this->assertion_type;
        if($lifetime==null) {
            $array['lifetime'] = $lifetime;
        }      
        return $array;
    }

    /**
     * Function that checks if and url is https or http
     * @param String $url
     * @return bool true if is http, false if it is https
     */
    private function isntHTTPS($url) {
        if($this->activeDebug) error_log("oauthClient.isHTTPS");
        return (false === stripos($url, "https://"));
    }

    /**
     *Function that makes bearer token requests and cryptographic token request
     *
     * Clients make bearer token requests by including the access token using the HTTP "Authorization"
     * request header with the "Token" authentication scheme as described in Section 5.1. The access
     * token is included using the "token" parameter.
     * For example, the client makes the following HTTPS request:
     * GET /resource HTTP/1.1 Host: server.example.com Authorization: Token token="vF9dft4qmT"
     *
     * When doing a cryptographic request (using an access token with a matching secret)
     *  the client uses the "token" attribute to include the access token in the request, and uses the
     *  "nonce", "timestamp", "algorithm", and "signature" attributes to apply the matching secret.
     * For example:
     * GET /resource HTTP/1.1 Host: server.example.com
     * Authorization: Token token="vF9dft4qmT",
     *                          nonce="s8djwd",
     *                          timestamp="137131200",
     *                          algorithm="hmac-sha256",
     *                          signature="wOJIO9A2W5mFwDgiDvZbTSMK/PY="
     *
     * @param String $bearer With the value"bearer" if it we want to do a bearer request, "hmac-sha256" if we want to do a cryptographic request
     * @param String $method Method that we choose for the resource request. These are the methods defined by the  draft-ietf-oauth-v2-05 in section 5.
     * The options are URI Query Parameter ("get"), HTTP Authorization request header ("header") and Form-Encoded Body Parameter ("form")
     * @param Array $extra Array of extra parameters added in case of necessity. Initialized by default to an empty array.
     */
    public function requestResource($bearer, $method, $extra=array()) {
        if($this->activeDebug) error_log("oauthClient.RequestResource");
        $this->bearer=$bearer;
        if(strcmp($bearer, "bearer")!=0) {
            $this->algorithm=$bearer;
            if(strcmp($method, "header")!=0) {
                if($this->activeDebug) error_log("Changing request resource method '".$method."' to 'header'");
                $method = "header";
            }
        }else if($this->isntHTTPS($this->url_rs)) {
            $this->algorithm="hmac-sha256";
            if($this->activeDebug) error_log("Changing algorithm to 'hmac-sha256' and url to 'https'. Clients SHOULD[7] request and utilize  an access token with a matching secret when making protected resource requests over an insecure channel");
            $this->url_rs = str_replace("http://", "https://", $this->url_rs);
            if(strcmp($method, "header")!=0) {
                if($this->activeDebug) error_log("Changing request resource method '".$method."' to 'header'");
                $method = "header";
            }
        }
        $data = $this->generateRequest($method);
        if(0==strcmp($method, "header")) {
            $this->headerResourceConnection($this->url_rs, $data, $extra);
        }else if(0==strcmp($method, "get")) {
            $this->getResourceConnection($this->url_rs, $data, $extra);
        }else if(0==strcmp($method, "form")) {
            $this->formResourceConnection($this->url_rs, $data, $extra);
        }else {
            $this->error = "Unknown request resource format";
        }
    }

    /**
     * Function that generates the resource request depending on the selected method and the type of token (bearer or access token).
     * @param <type> $method Method that we choose for the resource request. These are the methods defined by the  draft-ietf-oauth-v2-05 in section 5.
     * The options are URI Query Parameter ("get"), HTTP Authorization request header ("header") and Form-Encoded Body Parameter ("form")
     * @return Array Resource Request.  If it is a bearer token request this array has got the parameter "token" and if it is an access token request, it has got the parameters "token", "nonce", "timestamp", "algorithm" and "signature"
     *
     */
    private function generateRequest($method) {
        if($this->activeDebug) error_log("oauthClient.generateRequest");
        $data=array("token"=>$this->access_token);
        if($this->algorithm!=null) {
            //1) Calculate the request timestamp and generate a request nonce
            // Timestamp: positive integer. current time. in seconds
            $timestamp = time();
            //Nonce: random string, unique
            $nonce =  $this->generateNonce($timestamp, $this->assertion, $this->expires_in);
            //2) Construct the normalize string =>   "timestamp,nonce,algorithm,method,hostname,requestresourceuri"
            if($this->isntHTTPS($this->url_rs)) {
                $port=$_SERVER['SERVER_PORT'];
            }else {
                $port=443;

            }
            $normString = $timestamp.",".$nonce.",".$this->algorithm.",".$method.",".$this->url_rs.":".$port.",".$this->scope;
            //3) Signature calculation => digest = hmac-sha256(key, text)
            $key = $this->key;
            $digest = hash_hmac("sha256", $normString, $key);
            $data["nonce"]=$nonce;
            $data["timestamp"]=$timestamp;
            $data["algorithm"]=$this->algorithm;
            $data["signature"]=$digest;

        }
        return $data;
    }

    /**
     * Makes the http post curl connection to requests the access token from the authorization server.
     * Stores the access token in the protected param 'access_token'. If an error occurs, it stores the error in the protected param 'error'.
     * @param <type> $url url to connect
     * @param <type> $data params to send
     */
    private function postConnection($url, $data) {
        if($this->activeDebug) error_log("oauthClient.postConnection");
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
        $output = curl_exec($ch);
        if(curl_errno($ch)) {
            if($this->activeDebug)  error_log( 'Curl error: ' . curl_error($ch) );
            $this->error = "Curl error";
        }else {
            $info= curl_getinfo($ch);
            if($info['http_code']==200) {
               if(json_decode($output)) {
                    $aux = (array)json_decode($output);
                    $this->access_token = $aux['access_token'];
                    $this->expires_in = $aux['expires_in'];
                }else {
                    $this->error="Authorization Server's response unknown.";
                }
            }else {
                $this->error = $output;
            }
        }
        curl_close($ch);
    }


    /**
     * Makes the http authorization header curl request to get the resource form the resource server.
     * Stores the resource in the protected param 'resource'. If an error occurs, it stores the error in the protected param 'error'.
     * @param String $url url to connect
     * @param Array $data - params to send
     * @param Array $extra Array of extra parameters added in case of necessity. Initialized by default to an empty array.
     */
    private function headerResourceConnection($url, $data, $extra) {
        if($this->activeDebug) error_log("oauthClient.headerConnection");
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLINFO_HEADER_OUT, 1);
        curl_setopt($ch, CURLOPT_HEADER, 1);
        $opt = 'Authorization:  Token token="'.$data['token'].'"';
        if($this->algorithm!=null) {
            $opt.= ',nonce="'.$data['nonce'].'",';
            $opt.= 'timestamp="'.$data['timestamp'].'",';
            $opt.= 'algorithm="'.$data['algorithm'].'",';
            $opt.= 'signature="'.$data['signature'].'"';
        }
        if(sizeof($extra)>0) {
            $i=1;
            foreach($extra as $key => $val) {
                if(strcmp($key,"token")==0) {
                    $key=$key.$i;
                    $i++;
                }
                $opt.= ','.$key.'="'.$val.'"';
            }
        }
        curl_setopt($ch, CURLOPT_HTTPHEADER, array($opt));
        $output = curl_exec($ch);
        if(curl_errno($ch)) {
            if($this->activeDebug) error_log( 'Curl error: ' . curl_error($ch) );
        }else {
            $info= curl_getinfo($ch);
            $this->manageCURLResponse($info, $output);
            curl_close($ch);
        }
    }


    /**
     * Manage the response of the curl exception. Stores the resource in the protected param 'resource'.
     * If an error occurs, it stores the error in the protected param 'error'.
     * @param String $url url to connect
     * @param Array $info  $info = curl_getinfo($ch);
     * @param String $output  Output of the curl execution.
     */
    private function manageCURLResponse($info, $output) {
        if($this->activeDebug) error_log("oauthClient.manageCURLResponse");
        if($info['http_code']==200) {
            $array = explode("Content-Type: text/html", $output);
            if(false!=$array && sizeof($array)>1) {
                $this->resource = $array[1]; 
            }else {
                $this->resource =$this->transformResponse($output);
            }
        }else {
            if(false!=($date = $this->extractCustomHeader("Date: ","WWW-Authenticate:", $output))) {
                $this->isExpired=true;
                $this->error="The token is expired";
            }else
            if(false!=($error = $this->extractCustomHeader("error='","'", $output))) {
                $this->error = $error; 
            }else{
                $this->error= $output;
            }

        }
    }
    /**
     * This function transform an output of the curl execution. This method has to be customized depending on the case.
     * @param String $output Response of the CURL execution. It contains the http header of the response.
     * @return mixed Formatted string
     * @todo
     */
    private function transformResponse($output) {
        if($this->activeDebug) error_log("oauthClient.transformResponse");
        $auxN = strpos($output,"<Response>");
        $auxN1 = substr($output,$auxN);
        return $auxN1;
    }

    /**
     * Function that extracts some params of a header with regular expressions
     * @param String $start Start of the expression
     * @param String $end End of the expression
     * @param String $header String of the header
     * @return bool|String false if doesn't exists the pattern in the header and if it exists, the string with the results of the reg.exp.
     */
    private function extractCustomHeader($start,$end,$header) {
        $pattern = '/'. $start .'(.*?)'. $end .'/';
        if (preg_match($pattern, $header, $result)) {
            return $result[1];
        } else {
            return false;
        }
    }

    /**
     * Makes the http GET request to get the resource form the resource server.
     * Stores the resource in the protected param 'resource'. If an error occurs, it stores the error in the protected param 'error'.
     * @param String $url url to connect
     * @param Array $data - params to send
     * @param Array $extra Array of extra parameters added in case of necessity. Initialized by default to an empty array.
     */
    private function getResourceConnection($url, $data, $extra) {
        if($this->activeDebug) error_log("oauthClient.getConnection");
        $ch = curl_init();
        $string = "?";
        $string.= 'oauth_token='.$data['token'];
        if(sizeof($extra)>0) {
            $i=1;
            foreach($extra as $key => $val) {
                if(strcmp($key,"token")==0) {
                    $key=$key.$i;
                    $i++;
                }
                $string.= '&oauth_'.$key.'='.$val;
            }
        }
        curl_setopt($ch, CURLOPT_URL, $url.$string);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        $output = curl_exec($ch);
        if(curl_errno($ch)) {
            if($this->activeDebug) error_log( 'Curl error: ' . curl_error($ch) );
        }else {
            $info= curl_getinfo($ch);
            $this->manageCURLResponse($info, $output);
            curl_close($ch);
        }
    }

    /**
     * Makes the http POST request to get the resource form the resource server.
     * Stores the resource in the protected param 'resource'. If an error occurs, it stores the error in the protected param 'error'.
     * @param String $url url to connect
     * @param Array $data - params to send
     * @param Array $extra Array of extra parameters added in case of necessity. Initialized by default to an empty array.
     */
    private function formResourceConnection($url, $data, $extra) {
        if($this->activeDebug) error_log("oauthClient.formConnection");
        $string= 'oauth_token='.$data['token'];
        if(sizeof($extra)>0) {
            $i=1;
            foreach($extra as $key => $val) {
                if(strcmp($key,"token")==0) {
                    $key=$key.$i;
                    $i++;
                }
                $string.= '&oauth_'.$key.'='.$val;
            }
        }
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_POSTFIELDS, $string);
        $output = curl_exec($ch);
        if(curl_errno($ch)) {
            if($this->activeDebug) error_log( 'Curl error: ' . curl_error($ch) );
        }else {
            $info= curl_getinfo($ch);
            $this->manageCURLResponse($info, $output);
            curl_close($ch);
        }
    }


    /**
     * Function that generate a random nonce and stores it in the lite bd to have an
     * @param int $timestamp
     * @param String $assertion Assertion provided.
     * @param int $expires_in Lifetime of the access token in seconds.
     * @return String nonce
     */
    private function generateNonce($timestamp, $assertion, $expires_in) {
        $bd = new liteBD();
        $s = null;
        $enc = false;
        while(!$enc) {
            $nonce = $this->rand_str();
            if(!$bd->existsNonce($nonce, $timestamp,$assertion, $expires_in)) {
                if($bd->storeNonces($nonce, $timestamp,$assertion, $expires_in)!=false) {
                    $enc=true;
                }
            }
        }
        return $nonce;
    }

    /**
     * Function that generate a random string
     * @param <type> $length Length of the random string
     * @param <type> $chars Character used for the alphabet of the string
     * @return String
     */
    private function rand_str($length = 32, $chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890') {
        $chars_length = (strlen($chars) - 1);
        $string = $chars{rand(0, $chars_length)};
        for ($i = 1; $i < $length; $i = strlen($string)) {
            $r = $chars{rand(0, $chars_length)};
            if ($r != $string{$i - 1}) $string .=  $r;
        }
        return $string;
    }

    /**
     * Function that activates the error_log messages
     */
    private function activeDebug() {
        $this->activeDebug=true;
    }

    /**
     * Function that deactivates the error_log messages
     */
    private function deactiveDebug() {
        $this->activeDebug=false;
    }



    private function transformArray($array){
        $aux = array();
            foreach ($array as $key => $value) {
                if(is_array($value)){
                    $key = str_replace(":", "_dos_puntos_", $key);
                    $aux[$key] = $this->transformArray($value);
                }else{
                   $key = str_replace(":", "_dos_puntos_", $key);
                   $value =  str_replace(":", "_dos_puntos_", $value);
                   $aux[$key] =$value;
                }
            }
            return $aux;           
    }

    /* Getters */

    /**
     * Return the error
     * @return String  if an error exists or null if it doesn't exist
     */
    public function getError() {
        return $this->error;
    }

    /**
     * Return the access token.
     * Only for debugging purposes
     * @return String  if an error exists or null if it doesn't exist
     */
    public function getAccessToken() {
        return $this->access_token;
    }

    /**
     * Return the expiration time of the token in seconds.
     * @return int
     */
    public function getExpiresIn() {
        return $this->expires_in;
    }

    /**
     * Return the required resource.
     * The format of the resources can be different in each implementation.
     * The custom transformation of the resource format could be done in the private method 'transformResponse'
     * @return String
     */
    public function getResource() {
        return $this->resource;
    }
    private function get_key() {
        return $this->key;
    }

    public function set_key($key) {
        $this->key = $key;
    }

}


class oauthConfig {

    protected $oauth_as;
    protected $oauth_server;
    protected $scope;
    protected $assertion_type;
    protected $lifetime;
    protected $response_format;
    protected $token_type;
    protected $format;
    protected $response;
    protected $error;
    protected $key;

    public function __construct() {
        $this->oauth_as = "https://test76.rediris.es/oauth2/oauth_as/tokenEndpoint.php";
        $this->oauth_server ="https://www.rediris.es/sir/api/oauth_server/serverEndpoint.php";
        //Scope of the resources
        $this->scope = "http://www.rediris.es/sir/api/sps_available.php";
        //Type of the assertion
        $this->assertion_type = "urn:sir";
        //Lifetime of the access or bearer token
        $this->lifetime = 3600;
        //Requesting the resource with an Access Token or an Bearer Token
        $this->token_type="bearer";
        //Format of the request of the resorce. It can be http header, GET or POST with x-www-urlenconde...
        $this->format = 'header';
        $this->error=null;
        $this->key="xxxxx";
    }
    /**
     * Function that makes all the authorization flow of OAuth.
     * @param String $assertion Assertion
     * @return String Return null if there is an error, the resource if everything goes ok.
     */
    public function startFlow($assertion) {
        $ret = null;
        $oauth = new oauthClient($this->assertion_type);
         $oauth->set_key($this->key);
        $oauth->requestAccessToken($assertion, $this->scope, $this->oauth_as, $this->oauth_server, $this->lifetime);
        $error = $oauth->getError();
        if($error==null) {
            $oauth->requestResource($this->token_type, $this->format);
            if(($error2 = $oauth->getError())!=null) {
                $this->error = $error2;
            }else {
                $ret = $oauth->getResource();
            }
        }else {
            $this->error = $error;
        }
        return $ret;
    }

    public function getError() {
        return $this->error;
    }


    public function get_oauth_as() {
        return $this->oauth_as;
    }

    public function set_oauth_as($oauth_as) {
        $this->oauth_as = $oauth_as;
    }

    public function get_oauth_server() {
        return $this->oauth_server;
    }

    public function set_oauth_server($oauth_server) {
        $this->oauth_server = $oauth_server;
    }

    public function get_scope() {
        return $this->scope;
    }

    public function set_scope($scope) {
        $this->scope = $scope;
    }

    public function get_assertion_type() {
        return $this->assertion_type;
    }

    public function set_assertion_type($assertion_type) {
        $this->assertion_type = $assertion_type;
    }

    public function get_lifetime() {
        return $this->lifetime;
    }

    public function set_lifetime($lifetime) {
        $this->lifetime = $lifetime;
    }


    public function get_token_type() {
        return $this->token_type;
    }

    public function set_token_type($token_type) {
        $this->token_type = $token_type;
    }

    public function get_format() {
        return $this->format;
    }

    public function set_format($format) {
        $this->format = $format;
    }

    public function get_key() {
        return $this->key;
    }

    public function set_key($key) {
        $this->key = $key;
    }




}

?>